"""
SC's equivalent of `NoneType`, used for "not defined" values.
"""


class SCUndefType:
    """
    SC's equivalent of `NoneType`, used for "not defined" values.

    The purpose of this is to allow variables to be "without a value" even when
    `None` can be a legitimate value (so, it cannot be used as a "value not
    set" marker.

    This class is a singleton (i.e., there can only ever be one instance of
    it).
    """

    _instance = None

    def __new__(cls):
        """
        Return instance of `SC_undef`, ensuring that there is always only one.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance


SC_undef = SCUndefType()



"""
Loader that grabs settings from environment variables.
"""

from typing import Any

from .base import SC_LoaderFromDict


class SC_EnvironLoader(SC_LoaderFromDict):
    """
    Loader that grabs settings from environment variables.
    """

    priority = -171923
    enabled = False
    name_case = str.upper

    @classmethod
    def get_source(cls) -> Any:
        """
        Return dictionary that with settings.
        """
        import os
        return os.environ



import unittest.mock

from settings_collector import (
    SettingsCollector, SC_WeirdBugError, SC_LoaderBase, SC_ConfigError,
)

from tests.utils import TestsBase


class TestBasicFunctionality(TestsBase):

    def test_weird_bugs(self):
        for what in ("root", "root.SC_Data.scopes"):
            with unittest.mock.patch(
                f"settings_collector.SettingsCollector.SC_Data.{what}",
                None,
            ):
                with self.assertRaises(SC_WeirdBugError):
                    SettingsCollector._get_scope("scope")

    def test_name_case(self):
        source_name = "sOmEnAmE"
        for name_case, is_ok, result in (
            (None, True, source_name),
            (str.lower, True, source_name.lower()),
            ("str.lower", False, SC_ConfigError),
        ):
            with unittest.mock.patch(
                "settings_collector.SC_LoaderBase.name_case",
                name_case,
            ):
                if is_ok:
                    self.assertEqual(
                        SC_LoaderBase._get_source_name(source_name),
                        result,
                    )
                else:
                    with self.assertRaises(result):
                        SC_LoaderBase._get_source_name(source_name)



from settings_collector import SettingsCollector, SC_Setting, sc_defaults

from tests.utils import TestsBase


class my_settings(SettingsCollector):

    foo = SC_Setting("food")
    bar = SC_Setting("bard")


class TestBasicFunctionality(TestsBase):

    def test_defaults(self):

        @sc_defaults(my_settings)
        def f(x, foo):
            return f"{x}:{foo}"

        self.assertEqual(f(17), "17:food")
        self.assertEqual(f(17, "afoot"), "17:afoot")
        self.assertEqual(f(17, foo="afoot"), "17:afoot")
        self.assertEqual(f(x=17, foo="afoot"), "17:afoot")

        self.assertEqual(f(17), "17:food")
        self.assertEqual(f(x=17), "17:food")
        self.assertEqual(f(17, foo="afoot"), "17:afoot")

    def test_defaults_with_func_defaults(self):

        @sc_defaults(my_settings)
        def f(x, y=17, foo="toe-foo"):
            return f"{x}:{y}:{foo}"

        self.assertEqual(f(13, 19), "13:19:toe-foo")
        self.assertEqual(f(13, 19, "afoot"), "13:19:afoot")
        self.assertEqual(f(13, 19, foo="afoot"), "13:19:afoot")
        self.assertEqual(f(13, y=19, foo="afoot"), "13:19:afoot")

        self.assertEqual(f(13), "13:17:toe-foo")
        self.assertEqual(f(x=13), "13:17:toe-foo")
        self.assertEqual(f(13, foo="afoot"), "13:17:afoot")

    def test_defaults_on_method(self):

        class Foo:

            @sc_defaults(my_settings)
            def f(self, x, foo):
                return f"{x}:{foo}"

        self.assertEqual(Foo().f(17), "17:food")
        self.assertEqual(Foo().f(17, "afoot"), "17:afoot")
        self.assertEqual(Foo().f(17, foo="afoot"), "17:afoot")
        self.assertEqual(Foo().f(x=17, foo="afoot"), "17:afoot")

        self.assertEqual(Foo().f(17), "17:food")
        self.assertEqual(Foo().f(x=17), "17:food")
        self.assertEqual(Foo().f(17, foo="afoot"), "17:afoot")

    def test_defaults_on_classmethod(self):

        class Foo:

            @classmethod
            @sc_defaults(my_settings)
            def f(cls, x, foo):
                return f"{x}:{foo}"

        self.assertEqual(Foo.f(17), "17:food")
        self.assertEqual(Foo.f(17, "afoot"), "17:afoot")
        self.assertEqual(Foo.f(17, foo="afoot"), "17:afoot")
        self.assertEqual(Foo.f(x=17, foo="afoot"), "17:afoot")

        self.assertEqual(Foo.f(17), "17:food")
        self.assertEqual(Foo.f(x=17), "17:food")
        self.assertEqual(Foo.f(17, foo="afoot"), "17:afoot")

    def test_defaults_on_staticmethod(self):

        class Foo:

            @staticmethod
            @sc_defaults(my_settings)
            def f(x, foo):
                return f"{x}:{foo}"

        self.assertEqual(Foo.f(17), "17:food")
        self.assertEqual(Foo.f(17, "afoot"), "17:afoot")
        self.assertEqual(Foo.f(17, foo="afoot"), "17:afoot")
        self.assertEqual(Foo.f(x=17, foo="afoot"), "17:afoot")

        self.assertEqual(Foo.f(17), "17:food")
        self.assertEqual(Foo.f(x=17), "17:food")
        self.assertEqual(Foo.f(17, foo="afoot"), "17:afoot")

    def test_defaults_on_init(self):

        class Foo:

            @sc_defaults(my_settings)
            def __init__(self, x, foo):
                self.x = x
                self.foo = foo

            def f(self):
                return f"{self.x}:{self.foo}"

        self.assertEqual(Foo(17).f(), "17:food")
        self.assertEqual(Foo(17, "afoot").f(), "17:afoot")
        self.assertEqual(Foo(17, foo="afoot").f(), "17:afoot")
        self.assertEqual(Foo(x=17, foo="afoot").f(), "17:afoot")

        self.assertEqual(Foo(17).f(), "17:food")
        self.assertEqual(Foo(x=17).f(), "17:food")
        self.assertEqual(Foo(17, foo="afoot").f(), "17:afoot")



"""
Exceptions specific for the settings collector.
"""


class SC_Exception(Exception):
    """
    Base exception class for all others to inherit.
    """


class SC_ConfigError(SC_Exception):
    """
    Exception raised when there is a configuration error.
    """


class SC_WeirdBugError(SC_Exception):
    """
    Exception for situations that should not happen in for-mypy-only checks.
    """

    def __init__(self, message):
        super().__init__(
            f"{message} (this shouldn't happen; this check is here only to"
            f" appease mypy",
        )

class SC_NotALoader(SC_Exception):
    """
    Exception raised when something that should be a loader is not one.
    """

    def __init__(self, name):
        super().__init__(f"{name} is not a settings loader")



"""
Base class for settings loading classes.
"""

from typing import Iterable, Any, Optional, Type, Callable

from ..exceptions import SC_ConfigError


class _SC_LoaderBaseMeta(type):
    """
    Metaclass for `SC_LoaderBase` used to auto-register each class.
    """

    def __new__(metacls, name, bases, namespace, **kwargs):
        # Create the class.
        result = super().__new__(metacls, name, bases, namespace, **kwargs)

        # Register that class.
        from ..manager import SC_LoadersManager
        SC_LoadersManager.register_loader(result)

        # Return the new class as one normally would in `__new__`.
        return result


class SC_LoaderBase(metaclass=_SC_LoaderBaseMeta):
    """
    Base class for settings loading classes.
    """

    # A tuple of exceptions that are not errors but indicate that there are no
    # settings available for this loader.
    no_settings_exceptions: tuple[Type[Exception], ...] = (ImportError,)

    # Higher number, higher priority (i.e., its values override those from the
    # lower priority loaders). Makes sense when `load_all` is `True`.
    # Set this to a negative number for low-priority sources (see
    # `env.SC_EnvironLoader` class) and to a positive number if you
    # want to ensure that some source is preferred. Most loaders should
    # probably stick to the default 0.
    priority: int = 0

    # Should this loader be used by default?
    # This was mostly added to disable `SC_EnvironLoader` by default while
    # still being able to enable it for testing purposes.
    enabled: bool = True

    # The case used for names of the settings in the given loader's framework.
    # This is a function that takes and returns a string.
    # For example, Django loader will use `str.upper` because configuration
    # values in Django are traditionally defined as upper-case strings.
    name_case: Optional[Callable[[str], str]] = None

    @classmethod
    def _get_source_name(cls, name: str) -> str:
        """
        Return the name of the setting `name` in the framework's config.
        """
        if cls.name_case is None:
            return name
        elif callable(cls.name_case):
            return cls.name_case(name)
        else:
            raise SC_ConfigError(
                f"name_case must be a callable, not a {type(cls.name_case)}",
            )

    @classmethod
    def get_settings(
        cls, prefix: str, settings_names: Iterable[str],
    ) -> Optional[dict[str, Any]]:
        """
        Return the relevant settings values in a dictionary.

        You probably don't want to override this, as it only calls
        `load_settings` and ignores the exceptions that are meant to happen if
        the settings does not exist.

        :param prefix: A prefix to be added to each name. For example, is some
            setting's name is `"bar"` and the prefix is `"foo_"` it is loaded
            from `"foo_bar"` and saved as `"bar"`.
        :param settings_names: A list of string names to of the variables to
            load.
        :return: The relevant settings values in a dictionary or `None` if it's
            failing.
        """
        if not cls.enabled:
            return None

        if not isinstance(settings_names, list):
            settings_names = list(settings_names)

        prefix = cls._get_source_name(prefix)

        try:
            result, success = cls.load_settings(prefix, settings_names)
            if not success:
                return None
        except Exception as e:
            if isinstance(e, cls. no_settings_exceptions):
                return None
            else:
                raise
        else:
            return result

    @classmethod
    def load_settings(
        cls, prefix: str, settings_names: list[str],
    ) -> tuple[dict[str, Any], bool]:
        """
        Return the relevant settings values in a dictionary.

        This is what you want to override in subclasses for specific settings.

        :param prefix: A prefix to be added to each name. For example, is some
            setting's name is `"bar"` and the prefix is `"foo_"` it is loaded
            from `"foo_bar"` and saved as `"bar"`.
        :param settings_names: A list of string names to of the variables to
            load.
        :raise NotImplementedError: This method needs to be overriden in
            subclasses, with the override raising some exceptions from
            `no_settings_exceptions` tuple in case the settings doesn't exist.
            If needed, redefine that tuple in your subclass to contain all the
            relevant exception classes.
        :return: A tuple containing
            1. relevant settings values in a dictionary, and
            2. a Boolean describing the success of the loading.
        """
        raise NotImplementedError(
            f"do not use {cls.__name__} directly (use a class that inherits it"
            f" and has `load_settings` properly defined)",
        )  # pragma: no cover


class SC_LoaderFromAttribs(SC_LoaderBase):
    """
    Base for settings loader classes that load settings as attributes.
    """

    @classmethod
    def get_source(cls):
        """
        Return object that has settings set as attributes.
        """
        raise NotImplementedError(
            f"do not use {cls.__name__} directly (use a class that inherits it"
            f" and has `get_source` properly defined)",
        )  # pragma: no cover

    @classmethod
    def load_settings(
        cls, prefix: str, settings_names: list[str],
    ) -> tuple[dict[str, Any], bool]:
        """
        Return the relevant settings values in a dictionary.

        This is what you want to override in subclasses for specific settingss.

        :param prefix: A prefix to be added to each name. For example, is some
            setting's name is `"bar"` and the prefix is `"foo_"` it is loaded
            from `"foo_bar"` and saved as `"bar"`.
        :param settings_names: A list of string names to of the variables to
            load.
        :raise NotImplementedError: This method needs to be overriden in
            subclasses, with the override raising some exceptions from
            `no_settings_exceptions` tuple in case the settings doesn't exist.
            If needed, redefine that tuple in your subclass to contain all the
            relevant exception classes.
        :return: A tuple containing
            1. relevant settings values in a dictionary, and
            2. a Boolean describing the success of the loading.
        """
        source = cls.get_source()
        result = dict()
        for name in settings_names:
            try:
                source_name = cls._get_source_name(name)
                result[name] = getattr(source, f"{prefix}{source_name}")
            except AttributeError:
                pass
        return result, True  # Always `True`; failures happen with imports


class SC_LoaderFromDict(SC_LoaderBase):
    """
    Base for settings loader classes that load settings from dictionaries.
    """

    @classmethod
    def get_source(cls):
        """
        Return dictionary that with settings.
        """
        raise NotImplementedError(
            f"do not use {cls.__name__} directly (use a class that inherits it"
            f" and has `get_source` properly defined)",
        )  # pragma: no cover

    @classmethod
    def load_settings(
        cls, prefix: str, settings_names: list[str],
    ) -> tuple[dict[str, Any], bool]:
        """
        Return the relevant settings values in a dictionary.

        This is what you want to override in subclasses for specific settingss.

        :param prefix: A prefix to be added to each name. For example, is some
            setting's name is `"bar"` and the prefix is `"foo_"` it is loaded
            from `"foo_bar"` and saved as `"bar"`.
        :param settings_names: A list of string names to of the variables to
            load.
        :raise NotImplementedError: This method needs to be overriden in
            subclasses, with the override raising some exceptions from
            `no_settings_exceptions` tuple in case the settings doesn't exist.
            If needed, redefine that tuple in your subclass to contain all the
            relevant exception classes.
        :return: A tuple containing
            1. relevant settings values in a dictionary; and
            2. a Boolean describing the success of the loading (success here
               means that at least one value was found and loaded).
        """
        source = cls.get_source()
        result = dict()
        for name in settings_names:
            try:
                source_name = cls._get_source_name(name)
                result[name] = source[f"{prefix}{source_name}"]
            except KeyError:
                pass
        return result, bool(result)



"""
Settings loader for Django.
"""

from typing import Any

from .base import SC_LoaderFromAttribs


class SC_DjangoLoader(SC_LoaderFromAttribs):
    """
    Settings loader for Django.
    """

    name_case = str.upper

    @classmethod
    def get_source(cls) -> Any:
        """
        Return object that has settings set as attributes.
        """
        # type: ignore
        from django.conf import settings  # type: ignore
        return settings  # pragma: no cover



import unittest.mock

from settings_collector import SettingsCollector

# WARNING: `tests.custom_loaders` must be imported even if you don't use
# anything from it. That gets the mock loaders created and registered.
from tests.custom_loaders import (
    MOCK_LOADER_SETTINGS, SC_MockLoader, MockLoaderException,
    SC_TestAttrLoader,
)
from tests.utils import TestsBase


class TestLoaderBase(TestsBase):

    def test_get_settings_enabled(self):
        for enabled, expected in ((True, MOCK_LOADER_SETTINGS), (False, None)):
            with unittest.mock.patch(
                "tests.custom_loaders.SC_MockLoader.enabled", enabled,
            ):
                result = SC_MockLoader.get_settings("", tuple())
            self.assertEqual(result, expected)

    @unittest.mock.patch("tests.custom_loaders.SC_MockLoader.enabled", True)
    def test_get_settings_unexpected_exception(self):
        with self.assertRaises(MockLoaderException):
            SC_MockLoader.get_settings("error", tuple())


class TestCustomAttrLoader(TestsBase):

    def setUp(self):
        class _sct_attr_settings:
            scT__fOO = "bard"
            scT__sCOPE__fOO = "barman"

        import settings_collector
        settings_collector._sct_attr_settings = _sct_attr_settings
        super().setUp()

    def tearDown(self):
        import settings_collector
        del settings_collector._sct_attr_settings
        super().tearDown()

    def test_custom_loader(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                prefix = "SCt"
                greedy_load = False
            defaults = {"Foo": "default"}

        self.assertEqual(my_settings.Foo, "bard")
        self.assertEqual(my_settings("Scope").Foo, "barman")

    def test_missing_attribute(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                prefix = "SCt"
                greedy_load = False
            defaults = {"Foo": "default"}

        self.assertEqual(
            SC_TestAttrLoader.load_settings("scT__", ["Foo", "Missing"]),
            ({"Foo": "bard"}, True),
        )


class TestCustomDictLoader(TestsBase):

    def setUp(self):
        _sct_dict_settings = dict(
            scT__fOO="bard",
            scT__sCOPE__fOO="barman",
        )
        import settings_collector
        settings_collector._sct_dict_settings = _sct_dict_settings
        super().setUp()

    def tearDown(self):
        import settings_collector
        del settings_collector._sct_dict_settings
        super().tearDown()

    def test_custom_loader(self):
        class my_settings(SettingsCollector):
            class SC_Config:
                prefix = "SCt"
                greedy_load = False
            defaults = {"Foo": "default"}

        self.assertEqual(my_settings.Foo, "bard")
        self.assertEqual(my_settings("Scope").Foo, "barman")



"""
Mock loaders.

This mimic a framework that uses inverse case names. So, if you want a setting
called `Foo` in scope `Scope` and with prefix `SCt`, the framework will know it
as `scT__sCOPE__fOO`. It's ridiculous, but the purpose here is just to test
custom loaders.
"""

from typing import Any

from settings_collector import (
    SC_LoaderFromAttribs, SC_LoaderFromDict, SC_LoaderBase,
)


MOCK_LOADER_SETTINGS = {"foo": "bar"}


class MockLoaderException(Exception):
    pass


class SC_MockLoader(SC_LoaderBase):

    enabled = False

    @classmethod
    def load_settings(
        cls, prefix: str, settings_names: list[str],
    ) -> tuple[dict[str, Any], bool]:
        if prefix:
            raise MockLoaderException(prefix)
        else:
            return MOCK_LOADER_SETTINGS, True


class SC_TestAttrLoader(SC_LoaderFromAttribs):

    name_case = str.swapcase

    @classmethod
    def get_source(cls) -> Any:
        """
        Return object that has settings set as attributes.
        """
        # type: ignore
        from settings_collector import _sct_attr_settings  # type: ignore
        return _sct_attr_settings


class SC_TestDictLoader(SC_LoaderFromDict):

    name_case = str.swapcase

    @classmethod
    def get_source(cls) -> Any:
        """
        Return object that has settings set as attributes.
        """
        # type: ignore
        from settings_collector import _sct_dict_settings  # type: ignore
        return _sct_dict_settings



"""
Decorator for populating default argument values from settings collector.
"""

from functools import wraps
import inspect
from typing import Callable

from .collector import SettingsCollector


def sc_defaults(settings_collector: SettingsCollector) -> Callable:
    """
    Return decorator for populating default values from settings collector.
    """
    def outer(f: Callable) -> Callable:
        @wraps(f)
        def inner(*args, **kwargs):
            f_kwargs = dict()
            args_names = sc_kwargs | set(kwargs)
            for arg_name in args_names:
                try:
                    # Try to grab assigned value from keyword arguments.
                    f_kwargs[arg_name] = kwargs[arg_name]
                except KeyError:
                    try:
                        # Try to grab the index of `arg_name` among positional
                        # arguments...
                        arg_idx = arg2idx[arg_name]
                        # ...and check that it indeed has a value.
                        assert arg_idx < len(args)
                    except (KeyError, AssertionError):
                        # If not, apply the default.
                        f_kwargs[arg_name] = getattr(
                            settings_collector, arg_name,
                        )
            return f(*args, **f_kwargs)

        # The names of all settings defined in `settings_collector`.
        settings_names = set(settings_collector.get_settings_names())
        # The names of those settings that exist as function's properly named
        # arguments without any defaults (function's defaults override the ones
        # from `settings_collector`, although it's unclear why would you define
        # them in both places).
        sc_kwargs: set[str] = {
            arg_name
            for arg_name, arg_data in inspect.signature(f).parameters.items()
            if (
                # It's one of `settings_collector`'s settings.
                arg_name in settings_names
                # It's a properly named argument (i.e., not `*args`,
                # `**kwargs`, or something similar).
                and any(
                    arg_data.kind == getattr(inspect._ParameterKind, kind)
                    for kind in (
                        "KEYWORD_ONLY", "POSITIONAL_ONLY",
                        "POSITIONAL_OR_KEYWORD",
                    )
                )
                # There is no default value given in the function's signature.
                and arg_data.default is inspect._empty
            )
        }
        # Grab indexes of positional arguments to detect one they are given a
        # value.
        arg2idx = {
            arg_name: idx
            for idx, (arg_name, arg_data) in enumerate(
                inspect.signature(f).parameters.items(),
            )
            if (
                any(
                    arg_data.kind == getattr(inspect._ParameterKind, kind)
                    for kind in ("POSITIONAL_ONLY", "POSITIONAL_OR_KEYWORD")
                )
            )
        }
        return inner
    return outer



"""
Testing functions for Settings Collector.
"""

from typing import Tuple, Any

from settings_collector import SC_LoadersManager
from settings_collector import SettingsCollector, SC_Setting


TV_KeysType = str | Tuple[tuple[str, ...], str]
TV_ValuesType = dict[str, Any]
TV_Type = dict[TV_KeysType, TV_ValuesType]
TEST_VALUES: TV_Type = {
    "JustAString": {"kwargs": dict(), "value": "covfefe"},
    "AnInteger": {"kwargs": dict(value_type=int), "value": 17},
    (("scope1", "scope2"), "JustAString"): {"value": "covfefe, but scoped"},
    (("scope1", "scope2"), "AnInteger"): {"value": 1719},
}


SCTest = type(
    "SCTest",
    (SettingsCollector,),
    {
        "SC_Config": type("SC_Config", (object,), {"prefix": "SCTest"}),
        **{
            tv_name: SC_Setting(  # type: ignore
                default_on_error=False, **tv_def["kwargs"],
            )
            for tv_name, tv_def in TEST_VALUES.items()
            if isinstance(tv_name, str)
        }
    },
)


class _TestItem:
    """
    Auxiliary class for analysing test values.
    """

    def __init__(self, test_name: TV_KeysType) -> None:
        sep = SCTest.SC_Config.sep  # type: ignore
        prefix = SCTest.SC_Config.prefix  # type: ignore
        scopes_path: tuple[str, ...]
        if isinstance(test_name, tuple):
            self.settings = SCTest(sep.join(test_name[0]))
            self.name = test_name[1]
            scopes_path = (*test_name[0], test_name[1])
        elif isinstance(test_name, str):
            self.settings = SCTest
            self.name = test_name
            scopes_path = (test_name,)
        else:
            raise ValueError(  # pragma: no cover
                f"invalid type for {repr(test_name)}; must be a string or"
                f" a tuple of strings",
            )
        self.test_name_str = sep.join(scopes_path)
        self.source_name = f"{prefix}{sep}{self.test_name_str}"


def sc_test_print_expected():
    """
    Print expected values to use in the framework's config.
    """
    for test_name, test_def in TEST_VALUES.items():
        test_item = _TestItem(test_name)
        print(f"{test_item.source_name} = {repr(test_def['value'])}")


def sc_test_run(verbose: bool = True):
    """
    Run tests for Settings Collector and print the results.

    :param verbose: If set to `True`, a report is printed for each test.
        Otherwise, only failed test are printed or a message that all have
        passed if none fail.
    """
    def last_loader_name() -> str:
        last_loader = SC_LoadersManager.last_successful_loader
        if last_loader:
            return last_loader.__name__
        else:
            # If we ever get here, it's a bug.
            return "<NO LOADER>"  # pragma: no cover

    def msg() -> str:
        # This needs to be called after the `getattr` line below, so it cannot
        # be a simple string variable prepared before the `try...except` block.
        return f"for {test_item.source_name} loaded from {last_loader_name()}"

    report: list[str] = list()
    for test_name, test_def in TEST_VALUES.items():
        test_item = _TestItem(test_name)
        try:
            settings_value = getattr(test_item.settings, test_item.name)
            assert settings_value == test_def["value"]
        except ValueError as e:
            report.append(f"Error getting value {msg()}: {e}")
        except TypeError as e:
            report.append(f"Error casting value {msg()}: {e}")
        except AssertionError:
            report.append(
                f"Invalid value received {msg()}:"
                f" {repr(settings_value)} != {repr(test_def['value'])}",
            )
        except Exception as e:  # pragma: no cover
            # It shouldn't be possible to end up here.
            report.append(f"Unexpected error {msg()}: {e}")
        else:
            if verbose:
                report.append(
                    f"Test for {repr(test_item.test_name_str)} passed using"
                    f" {last_loader_name()}.",
                )

    if not report:
        report = ["Success: all tests have passed."]

    print("\n".join(report))



"""
Loader that grabs settings from Bottle's config.

Note: I never used Bottle. This was made based on reading the docs. Ping me
if it's wrong, please.
"""

from typing import Any

from .base import SC_LoaderFromDict


class SC_BottleLoader(SC_LoaderFromDict):
    """
    Loader that grabs settings from Bottle's config.
    """

    @classmethod
    def get_source(cls) -> Any:
        """
        Return dictionary that with settings.
        """
        import bottle  # type: ignore
        return bottle.default_app().config  # pragma: no cover



from .version import __version__  # noqa: W0611

from .loaders.base import (  # noqa: W0611
    SC_LoaderBase, SC_LoaderFromAttribs, SC_LoaderFromDict,
)
from .collector import SettingsCollector  # noqa: W0611
from .defaults import sc_defaults  # noqa: W0611
from .exceptions import (  # noqa: W0611
    SC_Exception, SC_ConfigError, SC_WeirdBugError, SC_NotALoader,
)
from .manager import SC_LoadersManager  # noqa: W0611
from .setting import SC_Setting  # noqa: W0611
from .value import SC_Value  # noqa: W0611
from .test import SCTest, sc_test_print_expected, sc_test_run  # noqa: W0611
from .undef import SC_undef  # noqa: W0611

from .loaders.base import SC_LoaderBase  # noqa: W0611
from .loaders.bottle import SC_BottleLoader  # noqa: W0611
from .loaders.cherrypy import SC_CherryPyLoader  # noqa: W0611
from .loaders.django import SC_DjangoLoader  # noqa: W0611
from .loaders.env import SC_EnvironLoader  # noqa: W0611
from .loaders.flask import SC_FlaskLoader  # noqa: W0611
from .loaders.pyramid import SC_PyramidLoader  # noqa: W0611
from .loaders.turbogears import SC_TurboGearsLoader  # noqa: W0611



from copy import copy, deepcopy

from settings_collector import SettingsCollector, SC_Setting

from tests.utils import TestsBase, patch_env


class TestSetting(TestsBase):
    """
    Setting of a value from the code (as opposed to getting it from congig).
    """

    def test_setting(self):
        class my_setting(SettingsCollector):
            foo = SC_Setting("bar")
            not_set = SC_Setting("not set")

        with patch_env(foo="rab", x__foo="beep"):
            self.assertEqual(my_setting.foo, "rab")
            self.assertEqual(my_setting("x").foo, "beep")
            my_setting.foo = "bra"
            my_setting("x").foo = "boop"
            my_setting("y").foo = "bapp"
            self.assertEqual(my_setting.foo, "bra")
            self.assertEqual(my_setting("x").foo, "boop")
            self.assertEqual(my_setting("y").foo, "bapp")

            my_setting.clear_cache()

            self.assertEqual(my_setting.foo, "rab")
            self.assertEqual(my_setting("x").foo, "beep")
            self.assertEqual(my_setting("y").foo, "rab")

            my_setting("a__b").not_set = "is set!"
            self.assertEqual(my_setting.not_set, "not set")
            self.assertEqual(my_setting("x").not_set, "not set")
            self.assertEqual(my_setting("y").not_set, "not set")
            self.assertEqual(my_setting("a").not_set, "not set")
            self.assertEqual(my_setting("b").not_set, "not set")
            self.assertEqual(my_setting("a__b").not_set, "is set!")
            self.assertEqual(my_setting("a__b__c").not_set, "is set!")

    def test_copy(self):
        sc1 = SC_Setting(default={"a": 17})
        sc2 = copy(sc1)
        self.assertIsNot(sc1, sc2)
        self.assertEqual(sc1.default, sc2.default)
        self.assertIs(sc1.default, sc2.default)

    def test_deepcopy(self):
        sc1 = SC_Setting(default={"a": 17})
        sc2 = deepcopy(sc1)
        self.assertIsNot(sc1, sc2)
        self.assertEqual(sc1.default, sc2.default)
        self.assertIsNot(sc1.default, sc2.default)
